# coding=utf-8


class DatastoreMixin(object):
    def __init__(self, org_id, volume_id):
        self.__org_id = org_id
        self.__volume_id = volume_id

    def _build_key(self, datastore, name):
        org_name = self.__org_id
        parent_key = datastore.key('DataVolume', self.__volume_id, namespace=org_name)

        return datastore.key(self._entity_kind, name, parent=parent_key, namespace=org_name)

    @property
    def _entity_kind(self):
        raise NotImplementedError(self._entity_kind)



# coding=utf-8
import binascii
import cPickle as pickle
import stat
import math
import zlib
from ...gae_datastore_mixin import GAEDatastoreMixin
from .tree_cache import TreeCache
from google.appengine.ext import ndb


class TreeCacheModel(ndb.Model):
    files_part = ndb.BlobProperty(required=True)
    sha = ndb.StringProperty(required=True)
    data_splits = ndb.IntegerProperty(required=True)
    volume_id = ndb.IntegerProperty()  # This is a new prop its needs to be required but for backward support it can't

    @classmethod
    def _get_kind(cls):
        return 'TreeCache'

    @property
    def name(self):
        entity_id = self.key.id()
        return '' if entity_id == 1 else entity_id

    @classmethod
    def key_with_data_part(cls, key, data_part):
        name = key.id()

        if data_part > 0:
            name = '%s:%s' % (name, data_part)

        return ndb.Key(cls, name, parent=key.parent(), namespace=key.namespace())

    @property
    def shred_keys(self):
        if self.data_splits == 1:
            return []

        return [self.key_with_data_part(self.key, part) for part in range(1, self.data_splits)]


class GAEDatastoreTreeCache(GAEDatastoreMixin, TreeCache):
    MAX_SHRED_SIZE = 1024 * 1024 - (10 * 1024)

    def __init__(self, org_id, volume_id):
        super(GAEDatastoreTreeCache, self).__init__(org_id, volume_id)

    def delete_all(self):
        self.delete_all_by_class(TreeCacheModel)

    @property
    def _entity_kind(self):
        return 'TreeCache'

    @classmethod
    def build_blob(cls, tree):
        for item in tree.iteritems():
            is_dir = item.mode & stat.S_IFDIR == stat.S_IFDIR
            yield (item.path.decode('utf8'), binascii.unhexlify(item.sha.decode('utf8')), is_dir)

    def _set_files(self, trees_objects):
        inserts = []
        for path, tree in trees_objects.items():
            name = path.decode('utf8')

            data_dump = pickle.dumps(list(self.build_blob(tree)), pickle.HIGHEST_PROTOCOL)

            data = zlib.compress(data_dump)

            data_splits = int(math.ceil(len(data) / float(self.MAX_SHRED_SIZE)))

            start_index = 0
            for part in range(data_splits):
                end_index = min(start_index + self.MAX_SHRED_SIZE, len(data))

                entity = TreeCacheModel(
                    key=TreeCacheModel.key_with_data_part(self._build_key(name), part),
                    data_splits=data_splits,
                    files_part=data[start_index:end_index],
                    volume_id=self._volume_id,
                    sha=tree.id.decode('utf8'))

                inserts.append(entity)

                start_index += self.MAX_SHRED_SIZE

        ndb.put_multi(inserts)

    def get_commit_id(self):
        entity = self._build_key('').get()

        if entity is None:
            return None

        return entity.sha

    @classmethod
    def _files_sha_array_to_dict(cls, files):
        result = {}
        for name, sha, is_dir in files:
            result[name] = (binascii.hexlify(sha), stat.S_IFDIR if is_dir else 0)

        return result

    @classmethod
    def __open_files_parts(cls, files_parts):
        one_big_string = ''
        for shred_entity in files_parts:
            if shred_entity is None:
                return None

            one_big_string += shred_entity.files_part

        return pickle.loads(zlib.decompress(one_big_string))

    @classmethod
    def __get_entities_and_shared(cls, keys):
        cache_entities = [entity for entity in ndb.get_multi(keys) if entity is not None]

        shared_keys = []
        for entity in cache_entities:
            shared_keys.extend(entity.shred_keys)

        shared_cache_entities = ndb.get_multi(shared_keys)

        return cache_entities, shared_cache_entities

    def _get_files(self, scans):
        keys = [self._build_key(path) for path, _, _ in scans]

        cache_entities, shared_cache_entities = self.__get_entities_and_shared(keys)

        files = {}
        shards_index = 0
        for entity in cache_entities:
            files_parts = [entity]
            files_parts.extend(shared_cache_entities[shards_index:shards_index + entity.data_splits - 1])
            shards_index += entity.data_splits
            files_data = self.__open_files_parts(files_parts)

            if files_data is not None:
                files[entity.name] = files_data

        return {name: self._files_sha_array_to_dict(data) for name, data in files.items()}



# coding=utf-8
import os

from mali_commands.data_volume import update_config_file
from tests_data_management.base import BaseTest
from mali_commands.utilities.os_utils import flatten_dir


class TestDataManagement(BaseTest):
    def testUpload(self):
        self.populate_data([
            'foo.txt',
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_file(volume_id, ['foo.txt'])

    def testUploadSingleProcess(self):
        self.populate_data([
            'foo.txt',
        ])

        params = {
            'object_store': {
                'use_multiprocess': False,
            }
        }

        volume_id = self.create_data_volume()

        update_config_file(volume_id, params)

        self.add_file(volume_id, ['foo.txt'])
        self.add_file(volume_id, ['foo.txt'])

    def testUploadDirectBucket(self):
        self.populate_data([
            'foo.txt',
        ])

        volume_id = self.create_data_volume()
        params = {
            'object_store': {
                'bucket_name': 'staging_data_volumes',
            }
        }

        update_config_file(volume_id, params)

        self.add_file(volume_id, ['foo.txt'])
        self.add_file(volume_id, ['foo.txt'])

    def testBasicFlow(self):
        """Test basic flow
        - Create a data volume
        - Add a file
        - Add a metadata
        - Commit
        - Clone
        """

        self.populate_data([
            'foo.txt',
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '1')
        self.assertEqual(result['total_meta_data'], '1')
        self.assertEqual(result['total_data_points'], '1')

        commit_id = result['commit_id']
        clone_path = os.path.join(self.clone_dir, '$phase/$name')
        self.clone(volume_id, clone_path, 'name:foo version:%s' % commit_id)
        flatten_dir(self.clone_dir)

        self.assertEqualDir(self.data_dir, self.clone_dir)

    def testMultipleCommits(self):
        self.populate_data([
            'foo.txt',
            'bar.txt'
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '1')
        self.assertEqual(result['total_meta_data'], '1')
        self.assertEqual(result['total_data_points'], '1')

        self.add_file(volume_id, ['bar.txt'])
        self.add_metadata(volume_id, ['bar.txt'], {'name': 'foo'})
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '1')
        self.assertEqual(result['total_meta_data'], '2')
        self.assertEqual(result['total_data_points'], '2')

        commit_id = result['commit_id']

        clone_path = os.path.join(self.clone_dir, '$phase/$name')
        self.clone(volume_id, clone_path, 'version:%s' % commit_id)
        flatten_dir(self.clone_dir)

        self.assertEqualDir(self.data_dir, self.clone_dir)

    def testMultipleCommitsNoMetadata(self):
        self.populate_data([
            'foo.txt',
            'bar.txt'
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '0')
        self.assertEqual(result['total_meta_data'], '0')
        self.assertEqual(result['total_data_points'], '1')

        self.add_file(volume_id, ['bar.txt'])
        result = self.commit(volume_id)

        self.assertEqual(result['version_total_data_points'], '1')
        self.assertEqual(result['version_total_meta_data'], '0')
        self.assertEqual(result['total_meta_data'], '0')
        self.assertEqual(result['total_data_points'], '2')

    def testModifyMeta(self):
        self.populate_data(['foo.txt'])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'state': 'ca'})
        commit_result1 = self.commit(volume_id)

        query_result1 = self.query(volume_id, 'version:%s state:ca' % commit_result1['commit_id'])

        self.assertEqual(1, len(query_result1))
        self.assertEqual('ca', query_result1[0]['state'])

        self.add_metadata(volume_id, ['foo.txt'], {'state': 'il'})
        commit_result2 = self.commit(volume_id)

        query_result2 = self.query(volume_id, 'version:%s state:il' % commit_result2['commit_id'])
        self.assertEqual(1, len(query_result2))
        self.assertEqual('il', query_result2[0]['state'])

        query_result2 = self.query(volume_id, 'version:%s state:ca' % commit_result2['commit_id'])
        self.assertEqual(0, len(query_result2))

    def testDuplicateMetaQueryStaging(self):
        self.populate_data(['foo.txt', 'bar.txt'])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt', 'bar.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'state': 'ca'})
        self.add_metadata(volume_id, ['foo.txt'], {'state': 'ca'})
        self.add_metadata(volume_id, ['bar.txt'], {'state': 'ca'})
        query_result1 = self.query(volume_id, 'version:staging state:ca')
        self.assertEqual(2, len(query_result1))
        self.assertEqual('ca', query_result1[0]['state'])
        self.assertEqual('ca', query_result1[1]['state'])

        self.assertEqual(sorted([item['path'] for item in query_result1]), ['bar.txt', 'foo.txt'])

    def testClone_metadataFromAPreviousCommit(self):
        self.populate_data(['foo.txt', 'bar.txt'])

        volume_id = self.create_data_volume()

        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        commit_result1 = self.commit(volume_id)
        commit_id1 = commit_result1['commit_id']

        self.add_file(volume_id, ['bar.txt'])
        self.add_metadata(volume_id, ['bar.txt'], {'name': 'bar'})
        result = self.commit(volume_id)
        commit_id = result['commit_id']

        query_result = self.query(volume_id, 'version:%s name:foo' % commit_id)
        self.assertEqual(len(query_result), 1)
        commit_id2 = result['commit_id']

        query_result = self.query(volume_id, 'version:%s name:foo' % commit_id2)
        query_result = query_result

    def testGroup(self):
        files = [
            '1.txt', '2.txt', '3.txt', '4.txt',
        ]
        self.populate_data(files)

        volume_id = self.create_data_volume()
        self.add_file(volume_id, files)
        self.add_metadata(volume_id, files, {'name': 'foo'})
        result = self.commit(volume_id)

        commit_id = result['commit_id']
        self.query(volume_id, 'name:foo group:name version:%s' % commit_id)



# coding=utf-8
from ...datastore_mixin import DatastoreMixin
from .tree_cache import TreeCache


class DatastoreTreeCache(DatastoreMixin, TreeCache):
    def __init__(self, connection):
        self.__connection = connection
        org_id = self.__connection.data_volume_config.org
        volume_id = self.__connection.data_volume_config.volume_id

        super(DatastoreTreeCache, self).__init__(org_id, volume_id)

    @property
    def _entity_kind(self):
        return 'TreeCache'

    def _set_files(self, trees_objects):
        with self.__connection.get_cursor() as datastore:
            inserts = []
            for path, tree in trees_objects.items():
                blob = '\n'.join(self.build_blob(tree))

                name = path.decode('utf8')

                entity = self.__connection.create_entity(self._build_key(datastore, name), exclude_from_indexes=('files', ))

                entity.update({
                    'files': blob.encode('utf8'),
                    'sha': tree.id.decode('utf8'),
                })

                inserts.append(entity)

            datastore.put_multi(inserts)

    def get_commit_id(self):
        with self.__connection.get_cursor() as datastore:
            key = self._build_key(datastore, '')
            entity = datastore.get(key)

            if entity is None:
                return None

            return entity['sha']

    def _get_files(self, scans):
        def name_from_key(entity):
            id_or_name = entity.key.name or entity.key.id
            return b'' if id_or_name == 1 else id_or_name

        with self.__connection.get_cursor() as datastore:
            keys = [self._build_key(datastore, path) for path, _, _ in scans]

            cache_entities = datastore.get_multi(keys)

            return {name_from_key(entity): self._files_sha_array_to_dict(entity['files']) for entity in cache_entities}



# coding=utf-8
import random
import string
import unittest

import httpretty
import logging


class Expando(object):
    def __init__(self, *args, **kw):
        self.__dict__.update(kw)


def global_some_random_shit(chars=string.ascii_lowercase + string.digits, size=8):
    return ''.join(random.choice(chars) for _ in range(size))


class di:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class BaseTest(unittest.TestCase):
    # class variable used by jmock style calls
    sharedController = None
    # so that nosetests run
    context = None

    def __init__(self, methodName=''):
        self.route_logging_stdout()

        super(BaseTest, self).__init__(methodName)
        self.datastore_probability = 1

    def setUp(self):
        super(BaseTest, self).setUp()

        httpretty.HTTPretty.allow_net_connect = False

        logging.debug('test setup')

    def tearDown(self):
        logging.debug('test tearDown')

        httpretty.HTTPretty.allow_net_connect = True

        super(BaseTest, self).tearDown()

    @classmethod
    def some_random_shit_number_int63(cls):
        return random.getrandbits(63)

    @classmethod
    def some_random_shit(cls, chars=string.ascii_lowercase + string.digits, size=8):
        return global_some_random_shit(chars, size)

    @classmethod
    def some_random_shit_alpha(cls, chars=string.ascii_lowercase, size=8):
        return global_some_random_shit(chars, size)

    @classmethod
    def some_random_sha_hash(cls):
        return cls.some_random_shit(chars=string.hexdigits, size=40)

    @classmethod
    def route_logging_stdout(cls):
        logging.basicConfig(level=logging.WARNING)
        # logging.getLogger().setLevel(logging.DEBUG)
        # import sys
        # logging.basicConfig(stream=sys.stdout)
        # logging.getLogger().setLevel(logging.DEBUG)

    @classmethod
    def get_obj(cls, **kwargs):
        return di(**kwargs)


class BaseCliTest(BaseTest):
    def setUp(self):
        from mali import add_commands

        super(BaseCliTest, self).setUp()

        add_commands()



# walk.py -- General implementation of walking commits and their contents.
# Copyright (C) 2010 Google, Inc.
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""General implementation of walking commits and their contents."""


from collections import defaultdict

import collections
import heapq
from itertools import chain

from dulwich.diff_tree import (
    RENAME_CHANGE_TYPES,
    tree_changes,
    tree_changes_for_merge,
    RenameDetector,
    )
from dulwich.errors import (
    MissingCommitError,
    )
from dulwich.objects import (
    Tag,
    )

ORDER_DATE = 'date'
ORDER_TOPO = 'topo'

ALL_ORDERS = (ORDER_DATE, ORDER_TOPO)

# Maximum number of commits to walk past a commit time boundary.
_MAX_EXTRA_COMMITS = 5


class WalkEntry(object):
    """Object encapsulating a single result from a walk."""

    def __init__(self, walker, commit):
        self.commit = commit
        self._store = walker.store
        self._get_parents = walker.get_parents
        self._changes = {}
        self._rename_detector = walker.rename_detector

    def changes(self, path_prefix=None):
        """Get the tree changes for this entry.

        :param path_prefix: Portion of the path in the repository to
            use to filter changes. Must be a directory name. Must be
            a full, valid, path reference (no partial names or wildcards).
        :return: For commits with up to one parent, a list of TreeChange
            objects; if the commit has no parents, these will be relative to
            the empty tree. For merge commits, a list of lists of TreeChange
            objects; see dulwich.diff.tree_changes_for_merge.
        """
        cached = self._changes.get(path_prefix)
        if cached is None:
            commit = self.commit
            if not self._get_parents(commit):
                changes_func = tree_changes
                parent = None
            elif len(self._get_parents(commit)) == 1:
                changes_func = tree_changes
                parent = self._store[self._get_parents(commit)[0]].tree
                if path_prefix:
                    mode, subtree_sha = parent.lookup_path(
                        self._store.__getitem__,
                        path_prefix,
                    )
                    parent = self._store[subtree_sha]
            else:
                changes_func = tree_changes_for_merge
                parent = [
                        self._store[p].tree for p in self._get_parents(commit)]
                if path_prefix:
                    parent_trees = [self._store[p] for p in parent]
                    parent = []
                    for p in parent_trees:
                        try:
                            mode, st = p.lookup_path(
                                self._store.__getitem__,
                                path_prefix,
                            )
                        except KeyError:
                            pass
                        else:
                            parent.append(st)
            commit_tree_sha = commit.tree
            if path_prefix:
                commit_tree = self._store[commit_tree_sha]
                mode, commit_tree_sha = commit_tree.lookup_path(
                    self._store.__getitem__,
                    path_prefix,
                )
            cached = list(changes_func(
              self._store, parent, commit_tree_sha,
              rename_detector=self._rename_detector))
            self._changes[path_prefix] = cached
        return self._changes[path_prefix]

    def __repr__(self):
        return '<WalkEntry commit=%s, changes=%r>' % (
          self.commit.id, self.changes())


class _CommitTimeQueue(object):
    """Priority queue of WalkEntry objects by commit time."""

    def __init__(self, walker):
        self._walker = walker
        self._store = walker.store
        self._get_parents = walker.get_parents
        self._excluded = walker.excluded
        self._pq = []
        self._pq_set = set()
        self._seen = set()
        self._done = set()
        self._min_time = walker.since
        self._last = None
        self._extra_commits_left = _MAX_EXTRA_COMMITS
        self._is_finished = False

        for commit_id in chain(walker.include, walker.excluded):
            self._push(commit_id)

    def _push(self, object_id):
        try:
            obj = self._store[object_id]
        except KeyError:
            raise MissingCommitError(object_id)
        if isinstance(obj, Tag):
            self._push(obj.object[1])
            return
        # TODO(jelmer): What to do about non-Commit and non-Tag objects?
        commit = obj
        if commit.id not in self._pq_set and commit.id not in self._done:
            heapq.heappush(self._pq, (-commit.commit_time, commit))
            self._pq_set.add(commit.id)
            self._seen.add(commit.id)

    def _exclude_parents(self, commit):
        excluded = self._excluded
        seen = self._seen
        todo = [commit]
        while todo:
            commit = todo.pop()
            for parent in self._get_parents(commit):
                if parent not in excluded and parent in seen:
                    # TODO: This is inefficient unless the object store does
                    # some caching (which DiskObjectStore currently does not).
                    # We could either add caching in this class or pass around
                    # parsed queue entry objects instead of commits.
                    todo.append(self._store[parent])
                excluded.add(parent)

    def next(self):
        if self._is_finished:
            return None
        while self._pq:
            _, commit = heapq.heappop(self._pq)
            sha = commit.id
            self._pq_set.remove(sha)
            if sha in self._done:
                continue
            self._done.add(sha)

            for parent_id in self._get_parents(commit):
                self._push(parent_id)

            reset_extra_commits = True
            is_excluded = sha in self._excluded
            if is_excluded:
                self._exclude_parents(commit)
                if self._pq and all(c.id in self._excluded
                                    for _, c in self._pq):
                    _, n = self._pq[0]
                    if self._last and n.commit_time >= self._last.commit_time:
                        # If the next commit is newer than the last one, we
                        # need to keep walking in case its parents (which we
                        # may not have seen yet) are excluded. This gives the
                        # excluded set a chance to "catch up" while the commit
                        # is still in the Walker's output queue.
                        reset_extra_commits = True
                    else:
                        reset_extra_commits = False

            if (self._min_time is not None and
                    commit.commit_time < self._min_time):
                # We want to stop walking at min_time, but commits at the
                # boundary may be out of order with respect to their parents.
                # So we walk _MAX_EXTRA_COMMITS more commits once we hit this
                # boundary.
                reset_extra_commits = False

            if reset_extra_commits:
                # We're not at a boundary, so reset the counter.
                self._extra_commits_left = _MAX_EXTRA_COMMITS
            else:
                self._extra_commits_left -= 1
                if not self._extra_commits_left:
                    break

            if not is_excluded:
                self._last = commit
                return WalkEntry(self._walker, commit)
        self._is_finished = True
        return None

    __next__ = next


class Walker(object):
    """Object for performing a walk of commits in a store.

    Walker objects are initialized with a store and other options and can then
    be treated as iterators of Commit objects.
    """

    def __init__(self, store, include, exclude=None, order=ORDER_DATE,
                 reverse=False, max_entries=None, paths=None,
                 rename_detector=None, follow=False, since=None, until=None,
                 get_parents=lambda commit: commit.parents,
                 queue_cls=_CommitTimeQueue):
        """Constructor.

        :param store: ObjectStore instance for looking up objects.
        :param include: Iterable of SHAs of commits to include along with their
            ancestors.
        :param exclude: Iterable of SHAs of commits to exclude along with their
            ancestors, overriding includes.
        :param order: ORDER_* constant specifying the order of results.
            Anything other than ORDER_DATE may result in O(n) memory usage.
        :param reverse: If True, reverse the order of output, requiring O(n)
            memory.
        :param max_entries: The maximum number of entries to yield, or None for
            no limit.
        :param paths: Iterable of file or subtree paths to show entries for.
        :param rename_detector: diff.RenameDetector object for detecting
            renames.
        :param follow: If True, follow path across renames/copies. Forces a
            default rename_detector.
        :param since: Timestamp to list commits after.
        :param until: Timestamp to list commits before.
        :param get_parents: Method to retrieve the parents of a commit
        :param queue_cls: A class to use for a queue of commits, supporting the
            iterator protocol. The constructor takes a single argument, the
            Walker.
        """
        # Note: when adding arguments to this method, please also update
        # dulwich.repo.BaseRepo.get_walker
        if order not in ALL_ORDERS:
            raise ValueError('Unknown walk order %s' % order)
        self.store = store
        if isinstance(include, bytes):
            # TODO(jelmer): Really, this should require a single type.
            # Print deprecation warning here?
            include = [include]
        self.include = include
        self.excluded = set(exclude or [])
        self.order = order
        self.reverse = reverse
        self.max_entries = max_entries
        self.paths = paths and set(paths) or None
        if follow and not rename_detector:
            rename_detector = RenameDetector(store)
        self.rename_detector = rename_detector
        self.get_parents = get_parents
        self.follow = follow
        self.since = since
        self.until = until

        self._num_entries = 0
        self._queue = queue_cls(self)
        self._out_queue = collections.deque()

    def _path_matches(self, changed_path):
        if changed_path is None:
            return False
        for followed_path in self.paths:
            if changed_path == followed_path:
                return True
            if (changed_path.startswith(followed_path) and
                    changed_path[len(followed_path)] == b'/'[0]):
                return True
        return False

    def _change_matches(self, change):
        if not change:
            return False

        old_path = change.old.path
        new_path = change.new.path
        if self._path_matches(new_path):
            if self.follow and change.type in RENAME_CHANGE_TYPES:
                self.paths.add(old_path)
                self.paths.remove(new_path)
            return True
        elif self._path_matches(old_path):
            return True
        return False

    def _should_return(self, entry):
        """Determine if a walk entry should be returned..

        :param entry: The WalkEntry to consider.
        :return: True if the WalkEntry should be returned by this walk, or
            False otherwise (e.g. if it doesn't match any requested paths).
        """
        commit = entry.commit
        if self.since is not None and commit.commit_time < self.since:
            return False
        if self.until is not None and commit.commit_time > self.until:
            return False
        if commit.id in self.excluded:
            return False

        if self.paths is None:
            return True

        if len(self.get_parents(commit)) > 1:
            for path_changes in entry.changes():
                # For merge commits, only include changes with conflicts for
                # this path. Since a rename conflict may include different
                # old.paths, we have to check all of them.
                for change in path_changes:
                    if self._change_matches(change):
                        return True
        else:
            for change in entry.changes():
                if self._change_matches(change):
                    return True
        return None

    def _next(self):
        max_entries = self.max_entries
        while max_entries is None or self._num_entries < max_entries:
            entry = next(self._queue)
            if entry is not None:
                self._out_queue.append(entry)
            if entry is None or len(self._out_queue) > _MAX_EXTRA_COMMITS:
                if not self._out_queue:
                    return None
                entry = self._out_queue.popleft()
                if self._should_return(entry):
                    self._num_entries += 1
                    return entry
        return None

    def _reorder(self, results):
        """Possibly reorder a results iterator.

        :param results: An iterator of WalkEntry objects, in the order returned
            from the queue_cls.
        :return: An iterator or list of WalkEntry objects, in the order
            required by the Walker.
        """
        if self.order == ORDER_TOPO:
            results = _topo_reorder(results, self.get_parents)
        if self.reverse:
            results = reversed(list(results))
        return results

    def __iter__(self):
        return iter(self._reorder(iter(self._next, None)))


def _topo_reorder(entries, get_parents=lambda commit: commit.parents):
    """Reorder an iterable of entries topologically.

    This works best assuming the entries are already in almost-topological
    order, e.g. in commit time order.

    :param entries: An iterable of WalkEntry objects.
    :param get_parents: Optional function for getting the parents of a commit.
    :return: iterator over WalkEntry objects from entries in FIFO order, except
        where a parent would be yielded before any of its children.
    """
    todo = collections.deque()
    pending = {}
    num_children = defaultdict(int)
    for entry in entries:
        todo.append(entry)
        for p in get_parents(entry.commit):
            num_children[p] += 1

    while todo:
        entry = todo.popleft()
        commit = entry.commit
        commit_id = commit.id
        if num_children[commit_id]:
            pending[commit_id] = entry
            continue
        for parent_id in get_parents(commit):
            num_children[parent_id] -= 1
            if not num_children[parent_id]:
                parent_entry = pending.pop(parent_id, None)
                if parent_entry:
                    todo.appendleft(parent_entry)
        yield entry



# coding=utf-8
from .base_connection import BaseConnection


class GAEDatastoreConnection(BaseConnection):
    def _create_connection(self, **kwargs):
        pass

    def _rollback(self):
        pass

    def _commit(self):
        pass

    def create_sql_helper(self):
        pass

    def _create_cursor(self):
        pass



# coding=utf-8
def get_dist(package):
    from pkg_resources import get_distribution

    return get_distribution(package)


def get_keywords(package):
    from pkg_resources import DistributionNotFound

    try:
        dist = get_dist(package)
    except DistributionNotFound:
        return None

    parsed_pkg_info = getattr(dist, '_parsed_pkg_info', None)

    if parsed_pkg_info is None:
        return None

    return parsed_pkg_info.get('Keywords')


def get_version(package):
    from pkg_resources import DistributionNotFound

    try:
        dist = get_dist(package)
    except DistributionNotFound:
        return None

    return str(dist.version)



# coding=utf-8
from ..connection_mixin import ConnectionMixin
from ..dulwich.refs import RefsContainer


SYMREF = b'ref: '


class SpannerRefContainer(RefsContainer, ConnectionMixin):
    def __init__(self, connection):
        super(SpannerRefContainer, self).__init__()
        self._create_table_if_needed()

    def get_packed_refs(self):
        return {}

    def read_loose_ref(self, name):
        raise NotImplementedError(self.read_loose_ref)

    def set_if_equals(self, name, old_ref, new_ref):
        raise NotImplementedError(self.set_if_equals)

    def add_if_new(self, name, ref):
        raise NotImplementedError(self.add_if_new)

    def remove_if_equals(self, name, old_ref):
        raise NotImplementedError(self.remove_if_equals)

    def _create_connection(self, **kwargs):
        raise NotImplementedError(self._create_connection)

    def _create_table_if_needed(self):
        raise NotImplementedError(self._create_table_if_needed)
